// PV227, Lesson 7: Particle systems, compute shaders, geometry shaders
#pragma once

#include "../../Framework/PV227.h"

using namespace std;
using namespace PV227;

//---------------------
//----    SCENE    ----
//---------------------

// Shader programs
ShaderProgram particle_simple_program;
ShaderProgram particle_textured_program;
ShaderProgram update_particles_program;
ShaderProgram update_particles_shared_program;

// A camera that allows us to look at the object from different views
SimpleCamera the_camera;

// Textures
GLuint particle_tex;

// Data of particles:
//		- current number of particles in simulation
//		- maximum number of particles for which the memory is allocated
int current_particle_count, max_particle_count;

//		- positions and velocities of all particles, in CPU. We need the position twice,
//			one for the previous frame, and one for the current frame, because the position of
//			all particles in the previous frame affects the position of particles in the new frame.
//			This is not true for the velocities, however, the velocities of other particles do not
//			affect velocity of one particle, so the velocities is present only once and they are
//			updated in place.
vector<glm::vec4> particle_positions[2], particle_velocities;

//		- positions and velocities of all particles, in OpenGL buffers
GLuint particle_positions_bo[2], particle_velocities_bo;

//		- colors of all particles, in OpenGL buffer. They do not change during the simulation.
GLuint particle_colors_bo;

//		- VAOs for rendering particles, one for each buffer with positions.
GLuint particle_vao[2];

//		- Index of the current position buffer to read from and to write into. The values are 0/1.
int current_read, current_write;

// Data of our camera - view matrix, projection matrix, etc.
CameraData_UBO CameraData_ubo;

// OpenGL query object to get time of update of one frame (when using compute shaders)
GLuint UpdateTimeQuery;

// Functions that works with scene objects
void reload_shaders();
void reset_particles();
void init_scene();
void update_scene(int app_time_diff_ms);
void render_scene();

//-------------------
//----    GUI    ----
//-------------------

// Main AntTweakBar object
TwBar *the_gui;

// Variables that are changed with GUI
bool use_geometry_shaders;		// What to use to render particles: geometry shaders / simple points
float particle_size;			// Size of one particle (in world/view space) when using gemoetry shaders
int reset_particles_seed;		// Seed to use when generating random positions of particles
int desired_particle_count;		// Number of particles to be used in simulation
float update_time_ms;
float accel_factor;				// Parameter of the simulation
float distance_threshold;		// Parameter of the simulation
float time_delta;				// Parameter of the simulation

// Place where the computation runs (CPU / GPU / optimized GPU), currently used and desired by the user
int current_update_method, desired_update_method;
const int UPDATE_ON_CPU = 0;
const int UPDATE_ON_GPU_SIMPLE = 1;
const int UPDATE_ON_GPU_SHARED = 2;

// Callbacks from the GUI
void TW_CALL reload(void *);
void TW_CALL reset(void *);

// Functions that works with GUI
void init_gui();

//---------------------------
//----    APPLICATION    ----

// Current window size
int win_width = 640;
int win_height = 480;

// Current time of the application in milliseconds, for animations
int app_time_ms = 0;
int last_glut_time = 0;
