#version 430

// Size of work group - 256 threads
layout (local_size_x = 256) in;

// Input data in Shader Storage Buffers
layout (std140, binding = 0) buffer PositionsInBuffer
{
	vec4 particle_positions_read[];
};
layout (std140, binding = 1) buffer PositionsOutBuffer
{
	vec4 particle_positions_write[];
};
layout (std140, binding = 2) buffer VelocitiesBuffer
{
	vec4 particle_velocities[];
};

// Input data in uniform variables
uniform int current_particle_count;
uniform float time_delta;
uniform float accel_factor;
uniform float distance_threshold;

void main()
{
	// Task 2: Implement n-body simulation on GPU. Each particle is computed by a separate thread
	//		Hint: Variables:
	//			particle_positions_read   .. array of vec4, positions of particles to read from
	//			particle_positions_write  .. array of vec4, positions of particles to write into
	//			particle_velocities       .. array of vec4, velocities of particles to read from and to write into
	//			current_particle_count    .. integer, number of active particles in simulation
	//			distance_threshold        .. float, objects closer than this value exert no force to each other
	//			accel_factor              .. a constant that affects the final acceleration
	//			time_delta                .. time elapsed since the last frame of our simulation
	//			gl_GlobalInvocationID.x   .. index of the current thread (global = references all running threads)

	vec3 position = vec3(particle_positions_read[gl_GlobalInvocationID.x]);
	vec3 velocity = vec3(particle_velocities[gl_GlobalInvocationID.x]);

	vec3 accel = vec3(0.0);

	for (int p = 0; p < current_particle_count; p++)
	{
		vec3 other = vec3(particle_positions_read[p]);
		vec3 direction = other - position;
		float dist_sq = dot(direction, direction);
		if (dist_sq > distance_threshold)
		{
			accel += normalize(direction) / dist_sq;
		}
	}

	accel *= accel_factor;
	position += velocity * time_delta + 0.5 * accel * time_delta * time_delta;
	velocity += accel * time_delta;

	particle_positions_write[gl_GlobalInvocationID.x] = vec4(position, 1.0);
	particle_velocities[gl_GlobalInvocationID.x] = vec4(velocity, 0.0);
}
