#version 430

// Size of work group - 256 threads
layout (local_size_x = 256) in;

// Input data in Shader Storage Buffers
layout (std140, binding = 0) buffer PositionsInBuffer
{
	vec4 particle_positions_read[];
};
layout (std140, binding = 1) buffer PositionsOutBuffer
{
	vec4 particle_positions_write[];
};
layout (std140, binding = 2) buffer VelocitiesBuffer
{
	vec4 particle_velocities[];
};

// Input data in uniform variables
uniform int current_particle_count;
uniform float time_delta;
uniform float accel_factor;
uniform float distance_threshold;

// Data shared between threads in work group - loaded positions
shared vec3 shared_positions[gl_WorkGroupSize.x];

void main()
{
	// Task 3: Copy here the result from nbody_compute.glsl, and then extend it with the use of shared memory.
	//		Hint: Variables:
	//			gl_LocalInvocationID.x    .. index of the current thread in its work group (local = references the threads in work group)
	//			gl_WorkGroupSize.x        .. size of a work group
	//			shared_positions          .. defined a few lines above, an array with data shared between threads in work group
	//			barrier()                 .. waits for other threads in work group, see the slides

	vec3 position = vec3(particle_positions_read[gl_GlobalInvocationID.x]);
	vec3 velocity = vec3(particle_velocities[gl_GlobalInvocationID.x]);

	vec3 accel = vec3(0.0);

	for (int p_group = 0; p_group < current_particle_count; p_group += int(gl_WorkGroupSize.x))
	{
		shared_positions[gl_LocalInvocationID.x] = vec3(particle_positions_read[p_group + gl_LocalInvocationID.x]);
		barrier();
		for (int p = 0; p < gl_WorkGroupSize.x; p++)
		{
			vec3 other = shared_positions[p];
			vec3 direction = other - position;
			float dist_sq = dot(direction, direction);
			if (dist_sq > distance_threshold)
			{
				accel += normalize(direction) / dist_sq;
			}
		}
		barrier();
	}

	accel *= accel_factor;
	position += velocity * time_delta + 0.5 * accel * time_delta * time_delta;
	velocity += accel * time_delta;

	particle_positions_write[gl_GlobalInvocationID.x] = vec4(position, 1.0);
	particle_velocities[gl_GlobalInvocationID.x] = vec4(velocity, 0.0);
}
